import { world, system, Player, EffectTypes, } from '@minecraft/server';
import ItemHandler from 'items/item.handler';
import MeleeWeapon from 'melee/melee-weapon';
import { initializeScoreboard } from 'utils/initialize.scoreboard';
export const MELEE_RIGHT_SCOREBOARD = 'melee_hit_right';
export const MELEE_LEFT_SCOREBOARD = 'melee_hit_left';
export const MELEE_RIGHT_COOLDOWN_SCOREBOARD = 'melee_hit_right_cooldown';
export const MELEE_LEFT_COOLDOWN_SCOREBOARD = 'melee_hit_left_cooldown';
export const MELEE_DRAW_SCOREBOARD = 'draw';
class MeleeEventsHandler {
    constructor() {
        this.onPlayerInteractWithEntityEvent = world.beforeEvents.playerInteractWithEntity;
        this.onItemStartUseEvent = world.afterEvents.itemStartUse;
        this.onItemStopUseEvent = world.afterEvents.itemStopUse;
        this.onPlayerLeaveEvent = world.beforeEvents.playerLeave;
        this.onWorldInitialize = world.afterEvents.worldInitialize;
        this.onHoldingEvent = system.runInterval(this.checkHolding.bind(this));
        this.onPlayerInteractWithEntityEvent.subscribe(this.onPlayerInteractWithEntity.bind(this));
        this.onItemStartUseEvent.subscribe(this.onItemStartUse.bind(this));
        this.onItemStopUseEvent.subscribe(this.onItemStopUse.bind(this));
        this.onPlayerLeaveEvent.subscribe(this.onPlayerLeave.bind(this));
        this.onWorldInitialize.subscribe(this.onWorldInitializeMethod.bind(this));
    }
    onWorldInitializeMethod() {
        initializeScoreboard(MELEE_RIGHT_SCOREBOARD);
        initializeScoreboard(MELEE_LEFT_SCOREBOARD);
        initializeScoreboard(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
        initializeScoreboard(MELEE_LEFT_COOLDOWN_SCOREBOARD);
        initializeScoreboard(MELEE_DRAW_SCOREBOARD);
        MeleeWeapon.importInstances();
    }
    onPlayerLeave(data) {
        const player = data.player;
        if (!player.isInit)
            return;
        delete player.weapon;
        system.run(() => {
            player.resetScore(MELEE_RIGHT_SCOREBOARD);
            player.resetScore(MELEE_LEFT_SCOREBOARD);
            player.resetScore(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
            player.resetScore(MELEE_LEFT_COOLDOWN_SCOREBOARD);
            player.resetScore(MELEE_DRAW_SCOREBOARD);
        });
    }
    onItemStartUse({ source, itemStack }) {
        if (source && !(source instanceof Player))
            return;
        if (!itemStack)
            return;
        const player = source;
        if (player.startedHoldingRightClickDate)
            return;
        if (!player.isInit)
            return;
        if (!player.isAbleToUse)
            return;
        for (const weapon of MeleeEventsHandler.MeleeSet) {
            if (!ItemHandler.isWeaponItem(itemStack, weapon))
                continue;
            if (!weapon.charges.enabled)
                continue;
            weapon.onStartHoldingRightClick(player);
        }
    }
    onItemStopUse({ source, itemStack }) {
        if (source && !(source instanceof Player))
            return;
        if (!itemStack)
            return;
        const player = source;
        if (!player.isInit)
            return;
        if (!player.isAbleToUse)
            return;
        if (player.startedHoldingRightClickDate === null)
            return;
        if (!player.weapon)
            return;
        for (const weapon of MeleeEventsHandler.MeleeSet) {
            if (!ItemHandler.isWeaponItem(itemStack, weapon))
                continue;
            weapon.onReleaseHoldingRightClick(player);
        }
    }
    checkHolding() {
        for (const p of world.getPlayers()) {
            const player = p;
            // console.warn(player.getRotation().x, player.getRotation().y)
            if (!player.isInit)
                continue;
            const inventory = player.getComponent('inventory');
            const item = inventory.container.getItem(player.selectedSlotIndex);
            if (player.selectedSlotIndex !== player.holdingWeaponSlot || !ItemHandler.isWeaponItemInstance(item)) {
                player.holdingWeaponSlot = player.selectedSlotIndex;
                player.ableToUseDate.reset();
                player.isAbleToUse = false;
                player.weapon = null;
                player.resetScore(MELEE_RIGHT_SCOREBOARD);
                player.resetScore(MELEE_LEFT_SCOREBOARD);
                player.resetScore(MELEE_RIGHT_COOLDOWN_SCOREBOARD);
                player.resetScore(MELEE_LEFT_COOLDOWN_SCOREBOARD);
                player.resetScore(MELEE_DRAW_SCOREBOARD);
                player.resetRightClickHolding();
                player.chargeIndex = null;
                delete player.weapon;
            }
            if (!ItemHandler.isWeaponItemInstance(item))
                continue;
            for (const weapon of MeleeEventsHandler.MeleeSet) {
                if (ItemHandler.isWeaponItem(item, weapon)) {
                    if (!player.weapon || weapon.itemTypeId !== player.weapon?.itemTypeId) {
                        player.weapon = weapon;
                        player.drawWeaponLastSlot = player.selectedSlotIndex;
                    }
                    player.addEffect(EffectTypes.get('weakness'), 1, {
                        showParticles: false,
                        amplifier: 255,
                    });
                    player.weapon.onHolding(player);
                    player.isHoldingWeapon = true;
                    player.holdingWeaponSlot = player.selectedSlotIndex;
                }
            }
        }
    }
    onPlayerInteractWithEntity(data) {
        const player = data.player;
        if (!player && !player.isValid())
            return;
        if (!player.weapon)
            return;
        if (!player.weapon?.rightAttack?.enabled)
            return;
        data.cancel = true;
    }
}
MeleeEventsHandler.MeleeSet = new Set();
export default MeleeEventsHandler;
