import { Vector3 } from 'utils/vectors';
import { system, } from '@minecraft/server';
import Particles from './particles/particles';
import { getBlockFromDirection } from 'utils/get-block-from-direction';
import { DamagesHandler } from './damages.handler';
import Sounds from './sounds/sounds';
export class SideDamgesHandler {
    constructor(player, attack) {
        this.player = player;
        this.attack = attack;
        this.touchedEntities = new Map();
        this.touchedBlocks = new Map();
        this.raysToSend = new Set();
        this.viewVector = new Vector3(this.player.getViewDirection().x, this.player.getViewDirection().y, this.player.getViewDirection().z);
        if (!this.attack.enabled)
            return;
        const lineCount = this.attack.sideAttack.lineCount + 1;
        const rayPerLine = this.attack.sideAttack.rayPerLine - 1;
        const linesYSpacingInBlocks = this.attack.sideAttack.areaHeight / lineCount;
        let currentHeight = this.attack.sideAttack.areaHeight / 2;
        for (let i = 0; i <= lineCount; i++) {
            const degreeSpacingBetweenRays = (this.attack.sideAttack.areaWeight.to - this.attack.sideAttack.areaWeight.from) / rayPerLine;
            for (let j = 0; j <= rayPerLine; j++) {
                const isMiddleOfTheLine = j >= (rayPerLine + 1) / 2;
                const l = isMiddleOfTheLine ? j - (rayPerLine + 1) : j;
                const angle = this.attack.sideAttack.areaWeight.from + j * degreeSpacingBetweenRays;
                const isMiddle = i >= lineCount / 2;
                const needToRemove = Math.abs((rayPerLine / (2 * (isMiddle ? i - lineCount : i))) * (this.attack.sideAttack.borderRadius * 0.01));
                if (l >= needToRemove || j === (rayPerLine + 1) / 2) {
                    const rightAngle = l ? angle : -angle;
                    const leftAngle = -rightAngle;
                    const rightRay = this.viewVector.increaseLength(this.attack.range).rotateVectorAroundY(rightAngle).add(new Vector3(0, currentHeight, 0));
                    this.raysToSend.add({
                        vector: rightRay,
                        angle: rightAngle,
                    });
                    const leftRay = this.viewVector.increaseLength(this.attack.range).rotateVectorAroundY(leftAngle).add(new Vector3(0, currentHeight, 0));
                    if (j !== rayPerLine / 2) {
                        this.raysToSend.add({
                            vector: leftRay,
                            angle: leftAngle,
                        });
                    }
                }
            }
            currentHeight -= linesYSpacingInBlocks;
        }
        this.sendRays();
    }
    sendRays() {
        this.raysToSend.forEach((ray) => {
            if (!this.attack.enabled)
                return;
            const entities = this.player.dimension
                .getEntitiesFromRay(this.player.getHeadLocation(), ray.vector, { maxDistance: this.attack.range })
                .filter((raycast) => raycast?.entity?.id !== this.player.id)
                .filter((raycast) => !DamagesHandler.skippedEntitiesIds.includes(raycast?.entity?.typeId));
            if (entities.length > 0) {
                entities.forEach((ent) => {
                    if (!ent.entity)
                        return;
                    if (!ent.entity.isValid())
                        return;
                    this.touchedEntities.set(ent.entity.id, { raycastHit: ent, angle: ray.angle });
                });
            }
            else {
                const blockRayCast = this.player.dimension.getBlockFromRay(this.player.getHeadLocation(), ray.vector, {
                    maxDistance: this.attack.range,
                });
                if (blockRayCast?.block && blockRayCast?.face)
                    this.touchedBlocks.set(JSON.stringify(blockRayCast.block.location), {
                        raycast: blockRayCast,
                        angle: ray.angle,
                    });
            }
            if (this.player.hasTag('debug')) {
                this.player.dimension.spawnParticle('minecraft:basic_flame_particle', {
                    x: this.player.getHeadLocation().x + ray.vector.x,
                    y: this.player.getHeadLocation().y + ray.vector.y,
                    z: this.player.getHeadLocation().z + ray.vector.z,
                });
            }
        });
    }
    applyDamages(damages) {
        let soundPlayed = false;
        this.touchedBlocks.forEach((blockRayCast) => {
            const hitTimeout = this.getHitTimeout(blockRayCast.angle);
            const block = getBlockFromDirection(blockRayCast.raycast.block, blockRayCast.raycast.face);
            system.runTimeout(() => {
                if (!this.attack.enabled)
                    return;
                Particles.spawn(block.location, block.dimension, this.attack.particles?.onHitBlock.particleIds);
                if (!soundPlayed) {
                    soundPlayed = true;
                    Sounds.play(this.attack.sounds?.onHitBlock, block.dimension, block.location);
                }
            }, hitTimeout);
        });
        const entitiesRaycastHits = this.touchedEntities.values();
        let damaged = false;
        for (const entityRaycast of entitiesRaycastHits) {
            const target = entityRaycast.raycastHit.entity;
            const hitTimeout = this.getHitTimeout(entityRaycast.angle);
            system.runTimeout(() => {
                if (!this.attack.enabled)
                    return;
                let attack;
                try {
                    attack = DamagesHandler.applyDamageToEntity(target, this.player, damages, this.attack.knockback, this.attack?.fire, {
                        sound: !damaged ? this.attack.sounds?.onHitEntity : null,
                        particles: this.attack.particles?.onHitEntity,
                        functions: this.attack.functions?.onHitEntity,
                    });
                }
                catch { }
                if (!damaged && attack && !this.attack?.disableItemDamage) {
                    damaged = true;
                    this.player.damageHandItem();
                }
            }, hitTimeout);
        }
    }
    getHitTimeout(angle /* in degree */) {
        if (!this.attack.enabled)
            return;
        let timeout;
        if (angle <= 0) {
            timeout = Math.abs(this.attack.sideAttack.areaWeight.from + Math.abs(angle));
        }
        else {
            timeout = this.attack.sideAttack.areaWeight.to + angle;
        }
        const maxTimeout = this.attack.sideAttack.areaWeight.to * 2;
        const divider = maxTimeout / ((this.attack.sideAttack.timeToSlash / 1000) * 20);
        timeout /= divider;
        return timeout;
    }
}
/* const view = player.getViewDirection()
const t = new Vector3(view.x, view.y, view.z).increaseLength(10);

function rotateVectorAroundY(vector, angle) {

    const angleRadians = angle * (Math.PI / 180) // Math.PI / 4; // 45 degrés en radians

    const cosTheta = Math.cos(angleRadians);
    const sinTheta = Math.sin(angleRadians);

    // Calculer les nouvelles composantes du vecteur après rotation autour de l'axe Y
    const newX = vector.x * cosTheta + vector.z * sinTheta;
    const newY = vector.y;
    const newZ = -vector.x * sinTheta + vector.z * cosTheta;

    // Retourner le nouveau vecteur avec les composantes mises à jour
    return new Vector3(newX, newY, newZ);
}

let i = -35

const id = system.runInterval(() => {
    if (i >= 35) system.clearRun(id)
    const tLeft = rotateVectorAroundY(t.increaseLength(10), i)
    player.dimension.spawnParticle('minecraft:basic_flame_particle', {
        x: player.getHeadLocation().x + tLeft.x,
        y: player.getHeadLocation().y + tLeft.y,
        z: player.getHeadLocation().z + tLeft.z,
    });

    player.dimension.getEntitiesFromRay(player.getHeadLocation(), tLeft).forEach(e => e?.entity?.applyDamage(10))

    i+= 5
}, 0)

let j = -35

const t1 = new Vector3(view.x, view.y, view.z).increaseLength(10);

const id1 = system.runInterval(() => {
    if (j >= 35) system.clearRun(id1)
        const tLeft = rotateVectorAroundY(t1.increaseLength(10), j).add(new Vector3(0, 2, 0))
    player.dimension.spawnParticle('minecraft:basic_flame_particle', {
        x: player.getHeadLocation().x + tLeft.x,
        y: player.getHeadLocation().y + tLeft.y,
        z: player.getHeadLocation().z + tLeft.z,
    });

    player.dimension.getEntitiesFromRay(player.getHeadLocation(), tLeft).forEach(e => e?.entity?.applyDamage(10))

    j+= 5
}, 0) */
