import CustomPlayer from 'player/custom-player';
import MeleeWeapon, { IAttack, IFire, IKnockback } from './melee-weapon';
import { Block, Entity, GameMode, EntityRaycastHit, system, world, EntityDamageCause } from '@minecraft/server';
import { getBlockFromDirection } from 'utils/get-block-from-direction';
import { distanceBetweenVectors } from 'utils/distanceBetweenVectors';
import { SideDamgesHandler } from './side-damages.handler';
import { getBlockCenteredLocation } from 'utils/get-block-centered-location';
import Particles, { IParticleParams } from './particles/particles';
import Sounds, { ISoundsParams } from './sounds/sounds';
import { Vector3 } from 'utils/vectors';
import { MELEE_RIGHT_SCOREBOARD } from 'events/melee.events.handler';
import { Functions, IAttackFunctions, TFunctions } from './functions';

export enum EAttackType {
	Right,
	Left,
}

export class DamagesHandler {
	private attack: IAttack;

	private attackVectors: Vector3[];

	public static skippedEntitiesIds: Array<Entity['id']> = ['minecraft:item', 'minecraft:xp_orb', 'minecraft:arrow', 'minecraft:item_frame'];

	// ATTACK CONTEXT
	private block: Block;
	private target: Entity;

	constructor(private weapon: MeleeWeapon, private player: CustomPlayer, private bypassDraw: boolean = false) {}
	private updateAttackCooldown = () => {
		if (!this.attack?.enabled) return;
		else this.player.ableToHitDate.setEnding(this.attack?.cooldown ? (this.attack.cooldown / 20) * 1000 : 0);
	};

	public setAttack(attack: IAttack) {
		if (!attack) return;
		if (!attack.enabled) return;

		this.attack = attack;
		this.attackVectors = [
			new Vector3(0, 0, 0), // default view Vector
			new Vector3(1, this.attack.precision, 0),
			new Vector3(0, this.attack.precision, 1),
			new Vector3(1, -this.attack.precision, 0),
			new Vector3(0, -this.attack.precision, 1),
		];

		this.player.lastAttack = this.attack;

		this.triggerAttack();
	}

	public setDamages(damages: number) {
		if (!this.attack) return;
		if (!this.attack.enabled) return;
		this.attack.damages = damages;
	}

	private getAttackContext(): void {
		if (!this.attack.enabled) return;

		const blockRaycast = this.player.getBlockFromViewDirection({ maxDistance: this.attack.range });

		const entitiesMap: Map<Entity['id'], EntityRaycastHit> = new Map();
		const playerViewDirection = this.player.getViewDirection();
		const viewVector = new Vector3(playerViewDirection.x, playerViewDirection.y, playerViewDirection.z);

		for (const vector of this.attackVectors) {
			const ray = viewVector.increaseLength(this.attack.range).add(vector);
			const entitiesRaycast = this.player.dimension.getEntitiesFromRay(this.player.getHeadLocation(), ray, { maxDistance: this.attack.range });

			entitiesRaycast
				.filter((raycast) => !DamagesHandler.skippedEntitiesIds.includes(raycast?.entity?.typeId))
				.forEach((raycast) => {
					if (raycast.entity && raycast.entity.id !== this.player.id) entitiesMap.set(raycast.entity.id, raycast);
				});
		}
		if (entitiesMap.size > 0) {
			this.target = [...entitiesMap.values()].sort((raycastA, raycastB) => raycastA.distance - raycastB.distance)[0]?.entity;
		} else if (blockRaycast?.block && blockRaycast?.face) this.block = getBlockFromDirection(blockRaycast.block, blockRaycast.face);
	}

	private triggerAttack(): void {
		if (!this.player.ableToHitDate.isFinished) return;
		if (!this.player.weapon) return;
		if (!this.bypassDraw && !this.player.isAbleToUse) return;
		if (!this.attack.enabled) return;

		this.updateAttackCooldown();

		system.runTimeout(() => {
			if (!this.player.isValid()) return;
			if (!this.player.weapon) return;
			if (!this.attack.enabled) return;

			this.getAttackContext();

			const sound: ISoundsParams = this.getSoundId();
			const particles: IParticleParams = this.getParticleId();
			const functions: TFunctions = this.getFunctions();

			if (this.attack.sideAttack) {
				new SideDamgesHandler(this.player, this.attack).applyDamages(this.attack.damages);
				return;
			}

			if (this.block) {
				if (distanceBetweenVectors(this.player.location, this.block.location) > this.attack.range) return;
				const location = getBlockCenteredLocation(this.block);
				Sounds.play(sound, this.player.dimension, location);
				Particles.spawn(location, this.block.dimension, particles?.particleIds);
			} else if (this.target) {
				if (!this.target.isValid()) return;
				if (distanceBetweenVectors(this.player.location, this.target.location) > this.attack.range) return;
				const attack = DamagesHandler.applyDamageToEntity(this.target, this.player, this.attack.damages, this.attack?.knockback, this.attack?.fire, {
					sound,
					particles,
					functions,
				});
				if (attack && !this.attack?.disableItemDamage) {
					this.player.damageHandItem();
				}
			} else Sounds.play(sound, this.player.dimension, this.player.getHeadLocation());
		}, this.attack.delay);
	}

	private getSoundId(): ISoundsParams {
		if (!this.attack.enabled) return;
		if (this.block) return this.attack?.sounds?.onHitBlock;
		else if (this.target) return this.attack?.sounds?.onHitEntity;
		else return this.attack?.sounds?.onMiss;
	}

	private getParticleId(): IParticleParams {
		if (!this.attack.enabled) return;
		if (this.block) return this.attack?.particles?.onHitBlock;
		else if (this.target) return this.attack?.particles?.onHitEntity;
	}

	private getFunctions(): TFunctions {
		if (!this.attack.enabled) return;
		else if (this.target) return this.attack?.functions?.onHitEntity;
	}

	public static applyDamageToEntity(
		target: Entity,
		player: CustomPlayer,
		damages: number,
		knockback: IKnockback,
		fire: IFire,
		effects?: {
			sound?: ISoundsParams;
			particles?: IParticleParams;
			functions?: TFunctions;
		}
	): boolean {
		const appliedDamage = target.applyDamage(damages, {
			cause: EntityDamageCause.entityAttack,
			damagingEntity: player,
		});

		const playerHit = target as CustomPlayer;

		if (appliedDamage) {
			if (effects.sound) Sounds.play(effects.sound, player.dimension, target.location);
			if (effects.particles) Particles.spawn(target.getHeadLocation(), target.dimension, effects.particles.particleIds);

			if (target.typeId == 'minecraft:player' && !playerHit.isAlive()) {
				world.sendMessage(`${player.name} ▬ι═ﺤ ${playerHit.name}`);
				player.runCommandAsync(`event entity @s kill_player`);
			}

			target.applyKnockback(player.getViewDirection().x, player.getViewDirection().z, knockback?.Ximpusle || 0.7, knockback?.Zimpusle || 0.3);

			if (effects?.functions?.length > 0) {
				Functions.executeMany(target, effects?.functions);
			}

			if (fire) {
				target.setOnFire(fire.time / 1000, fire.useEffects);
			}
		}

		return appliedDamage;
	}
}
