import CustomPlayer from 'player/custom-player';
import Sounds, { IAttackSounds, IOtherSounds, ISoundsParams } from 'melee/sounds/sounds';
import MeleeEventsHandler, {
	MELEE_DRAW_SCOREBOARD,
	MELEE_LEFT_COOLDOWN_SCOREBOARD,
	MELEE_LEFT_SCOREBOARD,
	MELEE_RIGHT_COOLDOWN_SCOREBOARD,
	MELEE_RIGHT_SCOREBOARD,
} from 'events/melee.events.handler';
import Particles, { IAttackParticles, IOtherParticles, IParticleParams } from './particles/particles';
import { EntityQueryOptions, EntityRaycastOptions, System, system, world, Player, Entity, HudElement, HudVisibility, Block } from '@minecraft/server';
import { LoadingBar } from 'utils/loading-bar';
import { Vector3 } from '../utils/vectors';
import { ChargeLoadingBar } from 'utils/charge-loading-bar';
import { SideDamgesHandler } from './side-damages.handler';
import { getBlockFromDirection } from 'utils/get-block-from-direction';
import { DamagesHandler, EAttackType } from './damages.handler';
// @ts-ignore
import { INSTANCES } from './instances/register';
import { IAttackFunctions } from './functions';

export interface IEnbaledOn {
	leftClick: boolean;
	rightClick: boolean;
}

export interface IKnockback {
	Ximpusle: number;
	Zimpusle: number;
}

export type IFire = {
	time: number; // in ms
	useEffects: boolean;
};

export type IAttack =
	| { enabled: false }
	| {
			enabled: true;
			displayCooldown?: boolean;
			fire?: IFire;
			knockback: IKnockback;
			delay: number; // in tick
			damages: number;
			disableItemDamage?: boolean;
			range: number; // in block
			cooldown?: number; // in tick
			precision: number; // in block
			sideAttack?: ISideDamagesSettings;
			spamCooldownEnabled?: boolean;
			sounds: IAttackSounds;
			particles: IAttackParticles;
			functions?: IAttackFunctions;
	  };

export type ICharge =
	| { enabled: false }
	| {
			enabled: true;
			timePerSquare: Array<number>;
			functionsTriggeredPerSquare?: Array<string>;
			functionsTriggeredAtReleasedPerSquare?: Array<string>;
			functionsTriggeredAtReleasedAfterDelayPerSquare?: Array<string>;
			displaySquares: boolean;
			triggerAttackIfLastSquareIsReached: boolean;
			damagePerSquare: Array<number>;
	  };

export interface ISideDamagesSettings {
	borderRadius: number; // in percent
	timeToSlash: number; // in ms
	rayPerLine: number;
	lineCount: number;
	areaHeight: number; // in blocks
	areaWeight: {
		from: number; // in degree
		to: number; // in degree
	};
}

export default class MeleeWeapon {
	public static readonly LOADING_BAR_TIME_PER_SQUARE: number = 250; // in ms

	public itemTypeId: string;

	// MAIN GUN PROPERTIES
	public name: string;

	public drawTime: number = 2500; // in ms
	public displayDrawCooldown: boolean = true;

	// HIT PARAMS
	public leftAttack?: IAttack = {
		enabled: true,
		damages: 10,
		delay: 40,
		cooldown: 500,
		range: 10,
		precision: 0.07,
		spamCooldownEnabled: true,
		knockback: {
			Ximpusle: 0.7,
			Zimpusle: 0.3,
		},
		sideAttack: {
			timeToSlash: 150, // in ms
			rayPerLine: 25,
			borderRadius: 50, // in percent
			lineCount: 25,
			areaHeight: 4.3, // in blocks
			areaWeight: {
				from: -25, // in degree
				to: 25, // in degree
			},
		},
		sounds: {
			onHitBlock: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
			onHitEntity: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
		},
		particles: {
			onHitBlock: {
				particleIds: ['minecraft:large_explosion'],
			},
			onHitEntity: {
				particleIds: ['minecraft:large_explosion'],
			},
		},
	};
	public rightAttack?: IAttack = {
		enabled: true,
		damages: 5,
		delay: 20,
		cooldown: 500,
		range: 10,
		precision: 0.07,
		knockback: {
			Ximpusle: 0.7,
			Zimpusle: 0.3,
		},
		sideAttack: {
			timeToSlash: 150, // in ms
			rayPerLine: 25,
			borderRadius: 50, // in percent
			lineCount: 25,
			areaHeight: 4.3, // in blocks
			areaWeight: {
				from: -25, // in degree
				to: 25, // in degree
			},
		},
		sounds: {
			onHitBlock: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
			onHitEntity: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
		},
		particles: {
			onHitBlock: {
				particleIds: ['minecraft:large_explosion'],
			},
			onHitEntity: {
				particleIds: ['minecraft:large_explosion'],
			},
		},
	};

	public chargeAttack?: IAttack = {
		enabled: true,
		damages: -1,
		delay: 0,
		cooldown: 0,
		range: 10,
		precision: 0.07,
		knockback: {
			Ximpusle: 0.7,
			Zimpusle: 0.3,
		},
		sideAttack: {
			timeToSlash: 150, // in ms
			rayPerLine: 25,
			borderRadius: 50, // in percent
			lineCount: 25,
			areaHeight: 4.3, // in blocks
			areaWeight: {
				from: -25, // in degree
				to: 25, // in degree
			},
		},
		sounds: {
			onHitBlock: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
			onHitEntity: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
		},
		particles: {
			onHitBlock: {
				particleIds: ['minecraft:large_explosion'],
			},
			onHitEntity: {
				particleIds: ['minecraft:large_explosion'],
			},
		},
	};
	// ATTACK on draw
	public initialAttack?: IAttack = {
		enabled: true,
		damages: 5,
		delay: 20,
		range: 10,
		precision: 0.07,
		knockback: {
			Ximpusle: 0.7,
			Zimpusle: 0.3,
		},
		sideAttack: {
			timeToSlash: 150, // in ms
			rayPerLine: 25,
			borderRadius: 50, // in percent
			lineCount: 25,
			areaHeight: 4.3, // in blocks
			areaWeight: {
				from: -25, // in degree
				to: 25, // in degree
			},
		},
		sounds: {
			onHitBlock: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
			onHitEntity: {
				soundIds: ['random.explode'],
				pitch: 2,
				volume: 100,
			},
		},
		particles: {
			onHitBlock: {
				particleIds: ['minecraft:large_explosion'],
			},
			onHitEntity: {
				particleIds: ['minecraft:large_explosion'],
			},
		},
	};

	// Charge System
	public charges: ICharge = {
		enabled: false,
	};

	// ANIMATIONS / SOUNDS
	public sounds: Record<keyof IOtherSounds, ISoundsParams> = {
		onDraw: {
			soundIds: ['random.explode'],
			pitch: 2,
			volume: 100,
		},
	};

	public particles: Record<keyof IOtherParticles, IParticleParams> = {
		onDraw: {
			particleIds: ['minecraft:large_explosion'],
		},
	};

	constructor() {
		MeleeEventsHandler.MeleeSet.add(this);
	}

	public onDraw(player: CustomPlayer): void {
		Sounds.playForPlayer(
			{
				...this.sounds.onDraw,
				location: player.getHeadLocation(),
			},
			player
		);
		Particles.spawn(player.getHeadLocation(), player.dimension, this.particles.onDraw.particleIds);

		if (this?.initialAttack?.enabled) {
			this.initialAttack.disableItemDamage = true
			const damageHandler = new DamagesHandler(this, player, true);
			damageHandler.setAttack(this?.initialAttack);
		}

		player.setScore(MELEE_LEFT_SCOREBOARD, 0);
		player.setScore(MELEE_RIGHT_SCOREBOARD, 0);
		player.setScore(MELEE_LEFT_COOLDOWN_SCOREBOARD, 0);
		player.setScore(MELEE_RIGHT_COOLDOWN_SCOREBOARD, 0);
		player.setScore(MELEE_DRAW_SCOREBOARD, 1);
		player.ableToUseDate.setEnding(this.drawTime);
	}

	public onHolding(player: CustomPlayer): void {
		player.updateActionBar();

		if (!player.isAbleToUse) {
			if (player.ableToUseDate.isFinished && player.ableToUseDate.running) {
				player.isAbleToUse = true;
				player.ableToUseDate.reset();
				player.setScore(MELEE_DRAW_SCOREBOARD, 0);
			} else if (!player.ableToUseDate.running) {
				this.onDraw(player);
				player.ableToUseDate.setEnding(this.drawTime);
			}
			return;
		}

		const damageHandler = new DamagesHandler(this, player);

		if (player.lastAttack && player.ableToHitDate.isFinished) {
			player.lastAttack = null;
		}

		if (player.getScore(MELEE_RIGHT_SCOREBOARD) === 1 && !this.charges.enabled) {
			world.scoreboard.getObjective(MELEE_RIGHT_COOLDOWN_SCOREBOARD)?.setScore(player, 1);
			world.scoreboard.getObjective(MELEE_RIGHT_SCOREBOARD)?.setScore(player, 0);
			if (this?.rightAttack?.enabled) {
				if (player.ableToHitDate.isFinished) {
					damageHandler.setAttack(this?.rightAttack);
				} else if (!player.ableToHitDate.isFinished && this.rightAttack?.spamCooldownEnabled) {
					player.ableToHitDate.setEnding((this.rightAttack.cooldown / 20) * 1000);
				}
			}
		}
		if (player.getScore(MELEE_LEFT_SCOREBOARD) === 1) {
			world.scoreboard.getObjective(MELEE_LEFT_COOLDOWN_SCOREBOARD)?.setScore(player, 1);
			world.scoreboard.getObjective(MELEE_LEFT_SCOREBOARD)?.setScore(player, 0);
			if (this?.leftAttack?.enabled) {
				if (player.ableToHitDate.isFinished) {
					damageHandler.setAttack(this?.leftAttack);
				} else if (!player.ableToHitDate.isFinished && this.leftAttack?.spamCooldownEnabled) {
					player.ableToHitDate.setEnding((this.leftAttack.cooldown / 20) * 1000);
				}
			}
		}

		if (player.ableToHitDate.isFinished) {
			world.scoreboard.getObjective(MELEE_LEFT_COOLDOWN_SCOREBOARD)?.setScore(player, 0);
			world.scoreboard.getObjective(MELEE_RIGHT_COOLDOWN_SCOREBOARD)?.setScore(player, 0);
		}
	}

	public onStartHoldingRightClick(player: CustomPlayer): void {
		if (player.startedHoldingRightClickDate) return;
		if (!this.charges.enabled) return;
		player.startedHoldingRightClickDate = Date.now();
		const chargeLoadingBar = new ChargeLoadingBar(this, player);
	}
	public onReleaseHoldingRightClick(player: CustomPlayer, clearInterval: boolean = true): void {
		if (!this.charges?.enabled) return;
		if (player.startedHoldingRightClickIntervalId && clearInterval) {
			system.clearRun(player.startedHoldingRightClickIntervalId);
			player.startedHoldingRightClickIntervalId = null;
			player.startedHoldingRightClickDate = null;
			world.scoreboard.getObjective(MELEE_RIGHT_SCOREBOARD)?.setScore(player, 0);
		}

		if (!player.ableToHitDate.isFinished) return;

		const damageIndex = player.chargeIndex;
		if (damageIndex == -1) return;

		if (this?.chargeAttack?.enabled) {
			const damage =
				player.chargeIndex > 0 ? this.charges.damagePerSquare.slice(0, player.chargeIndex).reduce((a, b) => a + b) : this.charges.damagePerSquare[0];
			const damageHandler = new DamagesHandler(this, player);
			damageHandler.setAttack(this?.chargeAttack);
			damageHandler.setDamages(damage);

			if (this.charges.enabled && this.charges.displaySquares)
				player.onScreenDisplay.setActionBar(`${' '.repeat(this.charges.timePerSquare.length).split('').join(' ')}`);

			if (this.charges?.functionsTriggeredAtReleasedPerSquare?.length > 0) {
				const functionName = this.charges.functionsTriggeredAtReleasedPerSquare[player.chargeIndex - 1];
				if (functionName) player.runCommandAsync(`function ${functionName}`);
			}

			system.runTimeout(() => {
				if (!this?.charges?.enabled) return;
				if (this.charges?.functionsTriggeredAtReleasedAfterDelayPerSquare?.length > 0) {
					const functionName = this.charges.functionsTriggeredAtReleasedAfterDelayPerSquare[player.chargeIndex - 1];
					if (functionName) player.runCommandAsync(`function ${functionName}`);
				}
				player.chargeIndex = null;
			}, this.chargeAttack?.delay);

			if (this.charges?.functionsTriggeredAtReleasedAfterDelayPerSquare?.length <= 0) player.chargeIndex = null;
		}
	}

	public static async importInstances(): Promise<void> {
		for (const weaponInstancesPath of INSTANCES) {
			const weapon = await import('./instances' + weaponInstancesPath);
			const builtWeapon = new weapon.default();
			MeleeEventsHandler.MeleeSet.add(builtWeapon);
		}
	}
}
