import {
	Container,
	Player,
	world,
	ContainerSlot,
	Entity,
	system,
	EntityHealthComponent,
	ItemStack,
	EnchantmentTypes,
	GameMode,
	ItemDurabilityComponent,
	ItemEnchantableComponent,
	EntityInventoryComponent,
	HudVisibility,
	HudElement,
} from '@minecraft/server';
import Timer from 'utils/date.timer';
import MeleeWeapon, { IAttack } from 'melee/melee-weapon';
import { LoadingBar } from 'utils/loading-bar';

export type TCustomEntity = Entity & {
	getHandItem(): ContainerSlot;
	getInventory(): Container;
};

// @ts-ignore
export default class CustomPlayer extends Player {
	public holdingWeaponSlot: number = null;
	public isHoldingWeapon: boolean = false;

	public isAbleToUse: boolean = false;
	public ableToUseDate: Timer;

	public weapon: MeleeWeapon | null = null;

	public drawWeaponLastSlot: number = null;

	public lastAttack: IAttack;
	public clearedActionBar: boolean;

	public startedHoldingRightClickDate: number = null;
	public startedHoldingRightClickIntervalId: number = null;
	public chargeIndex: number;

	public ableToLeftClickHitDate: Timer;
	public ableToRightClickHitDate: Timer;

	public ableToHitDate: Timer;

	public bypassClickLeftHitDate: boolean = false;
	public bypassClickRightHitDate: boolean = false;

	public isInit: boolean = false;

	public init(): void {
		this.ableToUseDate = new Timer();
		this.ableToLeftClickHitDate = new Timer();
		this.ableToRightClickHitDate = new Timer();
		this.ableToHitDate = new Timer();
		this.isInit = true;
	}

	public isAlive(): boolean {
		const health = this.getComponent('health') as EntityHealthComponent;
		return health.currentValue > 0;
	}

	public getInventory(): Container {
		const inventoryComponent = this.getComponent('inventory') as EntityInventoryComponent;
		return inventoryComponent.container;
	}

	public getHandItem(): ItemStack {
		return this.getInventory().getItem(this.selectedSlotIndex);
	}

	public damageHandItem(damages: number = 1, bypassCreative: boolean = false): void {
		if (!bypassCreative && this.getGameMode() === GameMode.creative) return;
		const itemStack = this.getHandItem();
		const durabilityComponent = itemStack.getComponent('durability') as ItemDurabilityComponent;
		const enchantementsComponent = itemStack.getComponent('enchantable') as ItemEnchantableComponent;

		const damageChance = durabilityComponent.getDamageChance(enchantementsComponent?.getEnchantment(EnchantmentTypes.get('unbreaking'))?.level || 0);

		const random = Math.random() * 100;
		if (random < damageChance) {
			durabilityComponent.damage += damages;
		}

		system.runTimeout(() => this.getInventory().setItem(this.selectedSlotIndex, itemStack), 2);
	}

	public getScore(objectiveName: string): number {
		if (!world.scoreboard.getObjective(objectiveName)?.hasParticipant(this)) return 0;
		return world.scoreboard.getObjective(objectiveName)?.getScore(this) || 0;
	}

	public resetScore(objectiveName: string): number {
		if (!world.scoreboard.getObjective(objectiveName)?.hasParticipant(this)) return 0;
		world.scoreboard.getObjective(objectiveName)?.removeParticipant(this);
	}
	public setScore(objectiveName: string, score: number): void {
		world.scoreboard.getObjective(objectiveName)?.setScore(this, score);
	}

	public resetRightClickHolding(): void {
		if (this.startedHoldingRightClickIntervalId) {
			system.clearRun(this.startedHoldingRightClickIntervalId);
			this.startedHoldingRightClickIntervalId = null;
			this.startedHoldingRightClickDate = null;
		}
	}

	public updateActionBar(): void {
		if (!this.weapon) return;

		if (!this.isAbleToUse && this.weapon.displayDrawCooldown) {
			this.clearedActionBar = false
			const loadingBar = new LoadingBar(
				MeleeWeapon.LOADING_BAR_TIME_PER_SQUARE,
				'█',
				{
					done: '§a',
					default: '§f',
				},
				this.weapon.drawTime,
				this.ableToUseDate.ending,
				Date.now()
			).create();

			this.onScreenDisplay.setActionBar(loadingBar);
		} else if (!this.ableToHitDate.isFinished && this.lastAttack && this.lastAttack.enabled && this.lastAttack.displayCooldown) {
			this.clearedActionBar = false
			const loadingBar = new LoadingBar(
				MeleeWeapon.LOADING_BAR_TIME_PER_SQUARE,
				'█',
				{
					done: '§a',
					default: '§f',
				},
				(this.lastAttack.cooldown * 1000) / 20,
				this.ableToHitDate.ending,
				Date.now()
			).create();

			this.onScreenDisplay.setActionBar(loadingBar);
		} else if (!this.clearedActionBar) {
			this.clearedActionBar = true
			this.onScreenDisplay.setActionBar(' ');
		}
	}
}

for (const key of Object.getOwnPropertyNames(CustomPlayer.prototype)) {
	Player.prototype[key] = CustomPlayer.prototype[key];
}

world.afterEvents.playerSpawn.subscribe((data) => {
	const player = data.player as CustomPlayer;
	player.init();
});

for (const player of world.getAllPlayers()) {
	const customPlayer = player as CustomPlayer;
	customPlayer.init();
}
